
data BTree a = Tip
             | Leaf a
             | Branch (BTree a) (BTree a)
             deriving Show

-- define a recursive function using "structural recursion"
-- define one pattern for each constructor in the BTree datatype
countLeaf :: BTree a -> Int
countLeaf Tip = 0
countLeaf (Leaf _) = 1
countLeaf (Branch t1 t2) = (countLeaf t1) + (countLeaf t2)

-- draw a picture of this tree
tree = Branch (Branch Tip (Leaf 5)) (Branch (Leaf 2) (Branch (Leaf 6) Tip))

n = countLeaf tree


-- define a recursive function using "structural recursion"
-- define one pattern for each constructor in the BTree datatype
sumLeaf :: (Num a) => BTree a -> a
sumLeaf Tip = 0
sumLeaf (Leaf i) = i
sumLeaf (Branch t1 t2) = (sumLeaf t1) + (sumLeaf t2)

m = sumLeaf tree


-- This functionn is a bit different. We don't know what
-- the maximum of an empty tree is, so we don't define the
-- function on the empty data constructor. However, there are an
-- infinite variety of empty trees (trees with no leafs), so this
-- function is not really well defined.
-- (In Haskell the maximum of an empty list is also not defined.)
maximumLeaf :: (Ord a) => BTree a -> a
maximumLeaf (Leaf i) = i
maximumLeaf (Branch t1  Tip) = maximumLeaf t1
maximumLeaf (Branch Tip t2)  = maximumLeaf t2
maximumLeaf (Branch t1  t2)  = max (maximumLeaf t1) (maximumLeaf t2)

p = maximumLeaf tree
q = maximumLeaf (Branch (Leaf 5) (Branch Tip Tip))

-- Try to define maximumLeaf this way. Return Nothing
-- when given a tree with no leafs.
maximumLeafM :: (Ord a) => BTree a -> Maybe a
maximumLeafM Tip = Nothing
maximumLeafM (Leaf i) = Just i
maximumLeafM (Branch t1 t2) = case r1 of
                                Nothing -> case r2 of
                                             Nothing -> Nothing
                                             Just j  -> Just j
                                Just i  -> case r2 of
                                             Nothing -> Just i
                                             Just j  -> Just (max i j)
                             where r1 = maximumLeafM t1
                                   r2 = maximumLeafM t2

p2 = maximumLeafM tree
q2 = maximumLeafM (Branch (Leaf 5) (Branch Tip Tip))
--q3 = maximumLeafM (Branch Tip (Branch Tip Tip))
